// Standard libraries and the CAPD library
#include <iostream>
#include <fstream>
#include <sstream>
#include <memory>
#include <chrono>
#include <cmath>
#include "capd/capdlib.h"
#include "capd/auxil/ConfigFileReader.h"
#include "capd/threading/ThreadPool.h"

// Algorithms specific for this program
#include "tictac.h"
#include "KSPoincareMap.h"
#include "embeddings.h"
#include "ProofData.h"

// Constants used in the computation.
const int g_order = 4;               /// Order of the PdeSolver.
const long double g_nu = 0.1212;     /// parameter of the KS PDE

const double g_setSize=1.91e-7;      /// scale-factor of the size of all HSets that appear in the computation
const double g_q = 1.5;              /// geometric decay
const int g_dissEncDimension = 11;   /// index of first variable on which dissipative Enclosure acts.
const int g_m = 15;                  /// dimension of HSets - see data structure HSet in the article
const double g_S = g_setSize*std::pow(g_q,g_m); /// constant S that gives geometrix bound on the tail in HSets. The same for all HSets used in the computation.

// #####################################################################

std::mutex poincareMapConstructorGuard;
std::mutex outputGuard;

/** This class implements an abstract class Task from the CAPD library. 
  * Objects of this type can be executed as tasks in a ThreadPool.
  * The method 'run' is executed asychrnonously in a thread from the thread pool.
  * 
  * ComputeImage computes rigorous bound for the Poincare map on a given domain.
  * The image of Poincare map is represented in an affine coordinate system.
*/
class ComputeImage : public threading::Task{
public:
  /// Task input data
  /// x+C*r - set on which we compute Poincare map
  const IVector x; const IMatrix C; const IVector r;
  /// y - > Q*(P^iterations(y)-u) - coordinate system in which result is represented
  const IMatrix Q; const IVector u;
  /// number of iterations of Poincare map
  const int iterations;
  /// Poincare section at u
  const IAffineSection section;
    
  /// Control data
  const int M;             /// number of coefficients computed by means of automatic differentiation
  const double tolerance;  /// acceptable error of PdeSolver per time step
  
  /// Task results
  interval rt;    /// bound for return time
  GeometricSeries result; /// bound for the result Q*(P^iterations(x+C*r)-u)
  /// Status 'Completed' means, that the rutine which computes Poincare map returned the result without throwing an exception - see method 'run' below.
  enum Validated { New, Completed };
  Validated status;

  ComputeImage(const ComputeImage& ) = delete;
  
  /// Constructor
  ComputeImage(
    IVector _x, IMatrix _C, IVector _r, 
    IMatrix _Q, IVector _u, 
    int _iterations, 
    IAffineSection _section,
    int _M, double _tolerance
  ) :
    x(_x), C(_C), r(_r), 
    Q(_Q), u(_u), 
    iterations(_iterations), 
    section(_section),
    M(_M), tolerance(_tolerance),
    result(_x.dimension(),0.,g_q),
    status(Validated::New)
  {}

  /// This method is executed asynchronously by a thread from the ThreadPool
  virtual void run(unsigned){
    std::unique_ptr<IKSPoincareMap> pm;
    try{
      {
        std::unique_lock<std::mutex> lock(poincareMapConstructorGuard);
        pm.reset(new IKSPoincareMap (g_dissEncDimension,M,g_nu,g_order,tolerance,g_q));
      }
      int D = x.dimension();
      pm->setSection(section);
      IVector x0(D), dr(D);
      split(r,x0,dr);
      /// here we call the routine which computes Poincare map. 
      this->result = (*pm)(x+C*x0,C,dr,Q,u,iterations,rt,g_S);
      /// we mark the status as completed.
      /// If the routine was able to integrate the set of initial conditions until its intersection with Poincare section (no exception caught)
      /// then we set status of computation as Completed. We will check if the computed bound satisfies required inequalities if all tasks finish computation.
      /// See class below CheckCoveringRelation
      this->status = Validated::Completed;
    }catch(std::exception& e){
      std::cout << "Exception: " << e.what() << std::endl;
    }
  }
};

// #####################################################################
/**
 * Let s1 be an HSet. This class is a task that asynchronously computes images of 
 * - left face of the HSet s1
 * - right face of the HSet s1
 * - entire set s1
 * 
 * After all tasks are completed, this class check if the inequalities required for covering relation are satisfied.
 */ 
class CheckCoveringRelation{
public:
  static auxil::ConfigFileReader reader;
  typedef std::vector<ComputeImage*> TaskGroup;

  CheckCoveringRelation(threading::ThreadPool& _pool, HSet* _s1, HSet* _s2, std::string _d)
    : pool(_pool), s1(_s1), s2(_s2), description(_d)
  {
    /// reader reads parameters of PdeSolver from the config file.
    /// These are: M = number of explicitly stored coorfinates in integration of PDE and tol=acceptable tolerance per one step of integration.
    /// See file data/config.dat
    int M = reader.get<int>(_d+".M");
    double tolerance = reader.get<double>(_d+".tol");

    /// C is the shape matrix of the set expressed in the coordinates on the section.
    /// and it is equal to product E*B, where E is the shape matrix of s1 and B are local orthogonal coordiantes on the section 
    IMatrix C = s1->E * s1->B;
    /// Q is the cooridinate system on the target section, in which the image wil be represented.
    /// If s1=s2 then Q=C^{-1} and we compute C^{-1}*P*C.
    /// Otherwise, we compute C2^{-1}*P*C1
    IMatrix Q = s2->inv_B * s2->inv_E;

    /// create task for computation of P(set)
    this->set = new ComputeImage(s1->c,C,s1->r,Q,s2->c,s1->iterations,s2->section,M,tolerance);
    
    /// create task for computation of P(rightFace(set))
    IVector r = s1->r;
    r[1] = r[1].rightBound();
    this->right = new ComputeImage(s1->c,C,r,Q,s2->c,s1->iterations,s2->section,M,tolerance);

    /// create task for computation of P(leftFace(set))
    r = s1->r;
    r[1] = r[1].leftBound();
    this->left = new ComputeImage(s1->c,C,r,Q,s2->c,s1->iterations,s2->section,M,tolerance);

    /// Start computation: execute method 'run' in a separate thread.
    this->t = new std::thread(&CheckCoveringRelation::run,std::ref(*this));
  }
  
  ~CheckCoveringRelation(){
    delete left;
    delete right;
    delete set;
  }

  void join(){
    if(t==nullptr)
      throw std::runtime_error("CheckCoveringRelation error: main thread has not been started!");
    t->join();
    delete t;
  }
  
  bool result() { return validated; }

private:
  /// This is the main function called asynchronously from the constructor
  void run(){
    this->start();    // start tasks that compute Poincare map on left, right face and the set
    this->wait();     // wait until they complete
    this->check();    // check inclusions for covering relation
    this->report();   // print report to stdout
  }

  /// start tasks that compute Poincare map
  void start(){
    pool.process(left);
    pool.process(right);
    pool.process(set);
  }
  
  /// auxiliary function, which checks if computed image 't' satisfies predicate 'condition'
  template<class Condition>
  static bool checkCondition(ComputeImage* t, Condition condition){
    return (t->status == ComputeImage::Validated::Completed) and condition(t->result);
  }

  /// GeometricSeries x has M-explicit coefficients, while infinite tail is bounded by S*q^{-i}.
  /// This function refines the constant S to the new dimension N<=M by bounding explicit coefficients N+1,...,M to the form new_S*q^{-i}
  static double getConstant(const GeometricSeries& x, int N){
    double S = x.getConstant().rightBound();
    for(int i=N;i<x.dimension();++i){
      interval t = abs(x[i])*power(interval(g_q),i);
      S = capd::max(S,t.rightBound());
    }
    return S;
  }

  /// This function checks inclusions for the covering relation s1=>s2
  bool check(){
    IVector r = s2->r; 

    /// lambda function (predicate), which checks if the projection of 'im' onto stable coordinates is inside the interior of r  
    auto accross = [r](const GeometricSeries& im){
      IVector pi_im = IVector(r.dimension()-2,im.getExplicitCoefficients().begin()+2);
      IVector pi_r = IVector(r.dimension()-2,r.begin()+2);
      /// Check inequality on finite number of coordinates and compare constants that give size of geometric tail.
      /// The constant in the image should be smaller than that of the set to be covered.
      return subsetInterior(pi_im,pi_r) and getConstant(im,r.dimension()) < g_S;
    };
    
    /// lambda function (predicate), which cheks if the projection onto unique unstable coordinate is ouside of r[1].
    auto mapaway = [r](const GeometricSeries& im){
      return abs(im[1])>abs(r[1]);
    };
    
    /// here we check the conditions for covering relation and return the result
    this->validatedLeft = checkCondition(left,mapaway);
    this->validatedRight = checkCondition(right,mapaway);
    this->validatedAcross = checkCondition(set,accross);
    /// Moreover two faces of s1 must be mapped into oposite sides of s2.r[1].
    /// It suffices to check that they have oposite signs.
    this->validatedToOpositeSides = left->result[1]*right->result[1] < 0.;
    this->validated = 
          this->validatedLeft and
          this->validatedRight and
          this->validatedAcross and
          this->validatedToOpositeSides;
    return this->validated;
  }
  
  /// This method prints to the standard output log from the computation and checking the inequalities
  void report(){
    std::ostringstream out;
    out << "############ " << description << " ################" << std::endl;
    if(this->validated)
      out << description << " checked" << std::endl;
    else
      out << description << " failed" << std::endl;
    /// projections onto Poincare section, i.e. we remove first irrelevant coordinate.
    int D = s2->c.dimension()-1;
    IVector imR(D,right->result.getExplicitCoefficients().begin()+1);
    IVector imL(D,left->result.getExplicitCoefficients().begin()+1);
    IVector imS(D,set->result.getExplicitCoefficients().begin()+1);
    out << "\n1. First (unstable) coordinate of right and left images must be mapped into oposite sides of the first coordinate of the box to be covered."
        << "\n2. All but first coordinates of the 'set image' must be mapped into corresponding coordinates of the box to be covered."
        << "\n3. The constant that gives uniform geometric bound on the tail in 'set image' must be smaller than that of target HSet."
        << "\n\nbox to be covered:\n" << IVector(D,s2->r.begin()+1)
        << "\n\nrightImage:\n" << imR
        << "\n\nleftImage:\n" << imL
        << "\n\nsetImage:\n"  << imS
        << "\n\nBoundOnTail(setImage)  : "  << getConstant(set->result,s2->r.dimension()) << " * " << g_q << "^{-i}"
        << "\nBoundOnTail(TargetHSet): "  << g_S  << " * " << g_q << "^{-i}"
        << "\n\nBound on return time: " << set->rt
        << std::endl;

    std::unique_lock<std::mutex> lock(outputGuard);
    std::cout << out.str() << std::flush;
  }
  
  void wait(){
    left->join();
    right->join();
    set->join();
  }

  threading::ThreadPool& pool;
  /// check covering between s1 and s2
  HSet *s1, *s2;
  /// text description of covering relation
  std::string description;
  bool validated = false;
  bool validatedLeft = false;
  bool validatedRight = false;
  bool validatedAcross = false;
  bool validatedToOpositeSides = false;
  std::thread* t = nullptr;
  
  ComputeImage *left, *right, *set;
};
auxil::ConfigFileReader CheckCoveringRelation::reader = auxil::ConfigFileReader("data/config.dat");

// ###############################################
/**
 * This function exports an HSet to the file - see description in index.html 
 */ 
void saveHSet(const HSet& h, const char* filename) {
  using namespace std;
  ofstream out(filename);
  out.precision(17);
  out << "S=" << g_S << endl;
  /// capd::vectalg::convertObject converts an interval vector or matrix (of diameter zero) to vector/matrix with double coefficients
  out << "a=" << vectalg::convertObject<DVector>(h.c) << endl;
  out << "E=" << vectalg::convertObject<DMatrix>(h.E) << endl;
  out << "B=" << vectalg::convertObject<DMatrix>(h.B) << endl;
  out << "r=" << h.r << endl;    
  out.close();
}

// ###############################################
/**
 * This auxiliary function reads the command line parameters,
 * and prepares data for rigorous computation. This means
 * - computation of coordinate systems at each h-set
 * - compuation of Poincare section at each h-set
 * The above computation are finite-dimensional and nonrigorous.
 * Eventually, the function exports all h-sets to the directory 'hsets'.
 */ 
int init(int argc,char* argv[]){
  LDKSPoincareMap pm(g_m,g_nu,8);
  int numberOfThreads = thread::hardware_concurrency();
  if(argc>1)  numberOfThreads = atoi(argv[1]);
  
  /// Init all h-sets --- see file ProofData.h.
  /// Here we read precomputed data from files, i.e.
  /// - approximate periodic points and heteroclinic points
  /// - coordinates at periodic points
  data.init(pm,g_m,g_setSize);

  /// Export all h-sets to text files. 
  saveHSet(*data.fp1,"hsets/N1.txt");
  saveHSet(*data.fp2,"hsets/N2.txt");
  for(int i=0;i<=10;++i){
    ostringstream filename;
    filename << "hsets/N" << i << "_12.txt";
    saveHSet(*data.h12[i],filename.str().c_str());
  }
  for(int i=0;i<=10;++i){
    ostringstream filename;
    filename << "hsets/N" << i << "_21.txt";
    saveHSet(*data.h21[i],filename.str().c_str());
  }

  return numberOfThreads;
}

// ###############################################

int main(int argc,char* argv[]){
  cout.precision(16);
  try{
    int numberOfThreads = init(argc,argv);
    threading::ThreadPool pool(numberOfThreads);

    ostringstream out;
    out << "Threads = " << numberOfThreads << endl; 
    out << "Dimension = " << g_m << endl;
    out << "Size of set = " << g_setSize << endl;
    out << "Order of PdeSolver = " << g_order << endl;

    /// Start clock
    auto tictac = tic(out.str().c_str());

      /// An array of tasks that validate covering relations
      vector<CheckCoveringRelation*> covrelList;

      /// Create and run tasks for validation of covering relations.

      /// Periodic points
      covrelList.push_back(new CheckCoveringRelation(pool,data.fp1,data.fp1,"N1->N1"));
      covrelList.push_back(new CheckCoveringRelation(pool,data.fp2,data.fp2,"N2->N2"));

      /// Heteroclinic chain a2->a1
      covrelList.push_back(new CheckCoveringRelation(pool, data.fp2,    data.h21[0], "N2->N0_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[0], data.h21[1], "N0_21->N1_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[1], data.h21[2], "N1_21->N2_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[2], data.h21[3], "N2_21->N3_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[3], data.h21[4], "N3_21->N4_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[4], data.h21[5], "N4_21->N5_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[5], data.h21[6], "N5_21->N6_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[6], data.h21[7], "N6_21->N7_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[7], data.h21[8], "N7_21->N8_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[8], data.h21[9], "N8_21->N9_21"));       
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[9], data.h21[10],"N9_21->N10_21"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h21[10],data.fp1,    "N10_21->N1"));

      /// Heteroclinic chain a1->a2
      HSet h = *data.fp1;
      h.iterations++;
      covrelList.push_back(new CheckCoveringRelation(pool, &h,          data.h12[0], "N1->N0_12")); 
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[0], data.h12[1], "N0_12->N1_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[1], data.h12[2], "N1_12->N2_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[2], data.h12[3], "N2_12->N3_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[3], data.h12[4], "N3_12->N4_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[4], data.h12[5], "N4_12->N5_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[5], data.h12[6], "N5_12->N6_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[6], data.h12[7], "N6_12->N7_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[7], data.h12[8], "N7_12->N8_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[8], data.h12[9], "N8_12->N9_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[9], data.h12[10],"N9_12->N10_12"));
      covrelList.push_back(new CheckCoveringRelation(pool, data.h12[10],data.fp2,    "N10_12->N2"));

      /// Wait for tasks completion
      for(CheckCoveringRelation* t : covrelList) t->join();

      /// Check and print the final result
      bool validated = true;
      for(CheckCoveringRelation* t : covrelList)
        validated = validated and t->result();
      cout << "\nFinal result: " << boolalpha << validated << endl;

    /// Stop clock and print time of computation
    tac(tictac);
    
    /// Release memory
    for(CheckCoveringRelation* t : covrelList) delete t;
    data.destroy();
  
  }catch(exception& e){
    cout << "Exception: " << e.what() << endl;
  }
}
